varying mediump vec4 var_frag_pos;
varying mediump vec3 var_normal;
varying mediump vec2 var_texcoord0;
varying mediump vec3 var_spherical_harmonic_lighting;

uniform mediump vec4 view_pos;
uniform mediump vec4 roughness;

uniform mediump vec4 directional_light_dir;
uniform mediump vec4 directional_light_ambient;
uniform mediump vec4 directional_light_diffuse;
uniform mediump vec4 directional_light_specular;

uniform mediump vec4 point_light_pos;
uniform mediump vec4 point_light_ambient;
uniform mediump vec4 point_light_diffuse;
uniform mediump vec4 point_light_specular;

uniform mediump vec4 spot_light_pos;
uniform mediump vec4 spot_light_ambient;
uniform mediump vec4 spot_light_diffuse;
uniform mediump vec4 spot_light_specular;
uniform mediump vec4 spot_light_inout_cut_off;  //x is cut off, y is outter cut off, z and w are unused.

uniform mediump vec4 light_attenuation_params;  //x is constant, y is linear, z is quadratic, w is unused.
uniform mediump vec4 tint;

uniform mediump sampler2D diffuse_map;

const float exposure = 0.8;

vec4 Albedo(vec2 uv)
{
    return texture2D(diffuse_map, uv, -1.0);
}

vec3 calculate_directional_light(vec3 dir, vec3 normal, vec3 view_dir, vec3 frag_color)
{
    vec3 light_dir = normalize(-dir);
    // diffuse shading
    float diff = max(dot(normal, light_dir), 0.0);
    // specular shading
    vec3 halfway_dir = normalize(light_dir + view_dir);
    float spec = pow(max(dot(normal, halfway_dir), 0.0), 64.0);
    // combine results
    vec3 ambient = vec3(directional_light_ambient) * frag_color;
    vec3 diffuse = vec3(directional_light_diffuse) * diff * frag_color * var_spherical_harmonic_lighting;
    vec3 specular = vec3(directional_light_specular) * spec * frag_color;
    return (ambient + diffuse + specular);
}

vec3 calculate_point_light(vec3 light_pos, vec3 normal, vec3 frag_pos, vec3 view_dir, vec3 frag_color)
{
    vec3 light_dir = normalize(light_pos - frag_pos);
    // diffuse shading
    float diff = max(dot(normal, light_dir), 0.0);
    // specular shading
    vec3 halfway_dir = normalize(light_dir + view_dir);
    float spec = pow(max(dot(normal, halfway_dir), 0.0), 64.0);
    // attenuation
    float distance = length(light_pos - frag_pos);
    float constant = light_attenuation_params.x;
    float linear = light_attenuation_params.y;
    float quadratic = light_attenuation_params.z;
    float attenuation = 1.0 / (constant + linear * distance + quadratic * (distance * distance));
    // combine results
    vec3 ambient = vec3(point_light_ambient) * frag_color;
    vec3 diffuse = vec3(point_light_diffuse) * diff * frag_color * var_spherical_harmonic_lighting;
    vec3 specular = vec3(point_light_specular) * spec * frag_color;
    ambient = ambient * attenuation;
    diffuse = diffuse * attenuation;
    specular = specular * attenuation;
    return (ambient + diffuse + specular);
}

vec3 calculate_spot_light(vec3 light_pos, vec3 normal, vec3 frag_pos, vec3 view_dir ,vec3 frag_color)
{
    vec3 light_dir = normalize(light_pos - frag_pos);
    // diffuse shading
    float diff = max(dot(normal, light_dir), 0.0);
    // specular shading
    vec3 halfway_dir = normalize(light_dir + view_dir);
    float spec = pow(max(dot(normal, halfway_dir), 0.0), 64.0);
    // attenuation
    float distance = sqrt(length(light_pos - frag_pos));
    float constant = light_attenuation_params.x;
    float linear = light_attenuation_params.y;
    float quadratic = light_attenuation_params.z;
    float attenuation = 1.0 / (constant + linear * distance + quadratic * (distance * distance));
    // spotlight intensitys
    float theta = dot(light_dir, normalize(vec3(light_pos.x, light_pos.y, light_pos.z)));
    float cut_off = spot_light_inout_cut_off.x;
    float outter_cut_off = spot_light_inout_cut_off.y;
    float epsilon = cut_off - outter_cut_off;
    float intensity = clamp((theta - outter_cut_off) / epsilon, 0.0, 1.0);
    // combine results
    vec3 ambient = vec3(spot_light_ambient) * frag_color * attenuation * intensity;
    vec3 diffuse = vec3(spot_light_diffuse) * diff * frag_color * attenuation * intensity * var_spherical_harmonic_lighting;
    vec3 specular = vec3(spot_light_specular) * spec * frag_color * attenuation * intensity;
    return (ambient + diffuse + specular);
}

void main()
{
    vec3 frag_color = vec3(Albedo(var_texcoord0.xy));
    vec3 V = normalize(vec3(view_pos - var_frag_pos));
    vec3 N = normalize(var_normal);
    
    vec3 color = calculate_directional_light(vec3(directional_light_dir), N, V, frag_color);

    color += calculate_spot_light(vec3(spot_light_pos), N, vec3(var_frag_pos), V, frag_color);

    //color = vec3(1.0) - exp(-color * exposure);

    //color = pow(color, vec3(1.0 / 2.2));

    gl_FragColor = vec4(color, 1.0) * tint;
}