#extension GL_OES_standard_derivatives : enable

varying highp vec4 var_frag_pos;
varying highp vec3 var_normal;
varying highp vec2 var_texcoord0;
varying highp vec3 var_spherical_harmonic_lighting;

uniform highp vec4 view_pos;
uniform highp vec4 roughness;

uniform highp vec4 directional_light_dir;
uniform highp vec4 directional_light_ambient;
uniform highp vec4 directional_light_diffuse;
uniform highp vec4 directional_light_specular;

uniform highp vec4 point_light_pos;
uniform highp vec4 point_light_ambient;
uniform highp vec4 point_light_diffuse;
uniform highp vec4 point_light_specular;

uniform highp vec4 spot_light_pos;
uniform highp vec4 spot_light_ambient;
uniform highp vec4 spot_light_diffuse;
uniform highp vec4 spot_light_specular;
uniform highp vec4 spot_light_inout_cut_off;  //x is cut off, y is outter cut off, z and w are unused.

uniform highp vec4 light_attenuation_params;  //x is constant, y is linear, z is quadratic, w is unused.
uniform highp vec4 tint;

uniform highp sampler2D diffuse_map;

const float exposure = 1.0;

vec4 Albedo(vec2 uv)
{
    vec2 resolution = vec2(2048, 2048);
    vec4 color1 = texture2D(diffuse_map, uv + (vec2(-1.0,-1.0)/resolution.xy), -1.0);
    vec4 color2 = texture2D(diffuse_map, uv + (vec2(1.0,-1.0)/resolution.xy), -1.0);
    vec4 color3 = texture2D(diffuse_map, uv + (vec2(-1.0,1.0)/resolution.xy), -1.0);
    vec4 color4 = texture2D(diffuse_map, uv + (vec2(1.0,1.0)/resolution.xy), -1.0);
    vec4 color5 = texture2D(diffuse_map, uv + (vec2(-1.0,0.0)/resolution.xy), -1.0);
    vec4 color6 = texture2D(diffuse_map, uv + (vec2(1.0,0.0)/resolution.xy), -1.0);
    vec4 color7 = texture2D(diffuse_map, uv + (vec2(0.0,-1.0)/resolution.xy), -1.0);
    vec4 color8 = texture2D(diffuse_map, uv + (vec2(0.0,1.0)/resolution.xy), -1.0);
    vec4 color = (color1 + color2 + color3 + color4 + color5 + color6 + color7 + color8) / 8.0;
    return color;
}

vec3 calculate_directional_light(vec3 dir, vec3 normal, vec3 view_dir, vec3 frag_color)
{
    vec3 light_dir = normalize(-dir);
    // diffuse shading
    float diff = max(dot(normal, light_dir), 0.0);
    // specular shading
    vec3 halfway_dir = normalize(light_dir + view_dir);
    float spec = pow(max(dot(normal, halfway_dir), 0.0), 32.0);
    // combine results
    vec3 ambient = vec3(directional_light_ambient) * frag_color;
    vec3 diffuse = vec3(directional_light_diffuse) * diff * frag_color * var_spherical_harmonic_lighting;
    vec3 specular = vec3(directional_light_specular) * spec * frag_color;
    return (ambient + diffuse + specular);
}

vec3 calculate_point_light(vec3 light_pos, vec3 normal, vec3 frag_pos, vec3 view_dir, vec3 frag_color)
{
    vec3 light_dir = normalize(light_pos - frag_pos);
    // diffuse shading
    float diff = max(dot(normal, light_dir), 0.0);
    // specular shading
    vec3 halfway_dir = normalize(light_dir + view_dir);
    float spec = pow(max(dot(normal, halfway_dir), 0.0), 32.0);
    // attenuation
    float distance = length(light_pos - frag_pos);
    float constant = light_attenuation_params.x;
    float linear = light_attenuation_params.y;
    float quadratic = light_attenuation_params.z;
    float attenuation = 1.0 / (constant + linear * distance + quadratic * (distance * distance));
    // combine results
    vec3 ambient = vec3(point_light_ambient) * frag_color;
    vec3 diffuse = vec3(point_light_diffuse) * diff * frag_color * var_spherical_harmonic_lighting;
    vec3 specular = vec3(point_light_specular) * spec * frag_color;
    ambient = ambient * attenuation;
    diffuse = diffuse * attenuation;
    specular = specular * attenuation;
    return (ambient + diffuse + specular);
}

vec3 calculate_spot_light(vec3 light_pos, vec3 normal, vec3 frag_pos, vec3 view_dir ,vec3 frag_color)
{
    vec3 light_dir = normalize(light_pos - frag_pos);
    // diffuse shading
    float diff = max(dot(normal, light_dir), 0.0);
    // specular shading
    vec3 halfway_dir = normalize(light_dir + view_dir);
    float spec = pow(max(dot(normal, halfway_dir), 0.0), 32.0);
    // attenuation
    float distance = length(light_pos - frag_pos);
    float constant = light_attenuation_params.x;
    float linear = light_attenuation_params.y;
    float quadratic = light_attenuation_params.z;
    float attenuation = 1.0 / (constant + linear * distance + quadratic * (distance * distance));
    // spotlight intensitys
    float theta = dot(light_dir, normalize(vec3(light_pos.x, light_pos.y, light_pos.z)));
    float cut_off = spot_light_inout_cut_off.x;
    float outter_cut_off = spot_light_inout_cut_off.y;
    float epsilon = cut_off - outter_cut_off;
    float intensity = clamp((theta - outter_cut_off) / epsilon, 0.0, 1.0);
    // combine results
    vec3 ambient = vec3(spot_light_ambient) * frag_color;
    vec3 diffuse = vec3(spot_light_diffuse) * diff * frag_color * var_spherical_harmonic_lighting;
    vec3 specular = vec3(spot_light_specular) * spec * frag_color;
    //var_spherical_harmonic_lighting
    ambient = ambient * attenuation * intensity;
    diffuse = diffuse * attenuation * intensity;
    specular = specular * attenuation * intensity;
    return (ambient + diffuse + specular);
}

void main()
{
    vec3 frag_color = vec3(Albedo(var_texcoord0.xy));

    vec3 V = normalize(vec3(view_pos - var_frag_pos));
   
    vec3 color = calculate_directional_light(vec3(directional_light_dir), normalize(var_normal), V, frag_color);

    color += calculate_point_light(vec3(point_light_pos), normalize(var_normal), vec3(var_frag_pos), V, frag_color);

    color += calculate_spot_light(vec3(spot_light_pos), normalize(var_normal), vec3(var_frag_pos), V, frag_color);

    //color = vec3(1.0) - exp(-color / exposure);

    //color = pow(color, vec3(1.0 / 2.2));

    gl_FragColor = vec4(color, 1.0) * tint;
}