varying mediump vec4 var_frag_pos;
varying mediump vec3 var_normal;
varying mediump vec2 var_texcoord0;
varying mediump vec3 var_spherical_harmonic_lighting;
varying mediump vec4 var_frag_pos_light_space;

uniform mediump vec4 view_pos;

uniform mediump vec4 directional_light_dir;
uniform mediump vec4 directional_light_ambient;
uniform mediump vec4 directional_light_diffuse;
uniform mediump vec4 directional_light_specular;

uniform mediump vec4 point_light_pos;
uniform mediump vec4 point_light_ambient;
uniform mediump vec4 point_light_diffuse;
uniform mediump vec4 point_light_specular;

uniform mediump vec4 spot_light_pos;
uniform mediump vec4 spot_light_ambient;
uniform mediump vec4 spot_light_diffuse;
uniform mediump vec4 spot_light_specular;
uniform mediump vec4 spot_light_inout_cut_off;  //x is cut off, y is outter cut off, z and w are unused.

uniform mediump vec4 light_attenuation_params;  //x is constant, y is linear, z is quadratic, w is unused.
uniform mediump vec4 tint;

uniform mediump sampler2D diffuse_map;
uniform mediump samplerCube env_map;
uniform mediump sampler2D depth_map;

const float near_plane = 0.1;
const float far_plane = 10.0;

vec2 GammaToLinearSpace (vec2 sRGB)
{
    return sRGB * (sRGB * (sRGB * 0.305306011 + 0.682171111) + 0.012522878);
}

vec3 GammaToLinearSpace (vec3 sRGB)
{
    return sRGB * (sRGB * (sRGB * 0.305306011 + 0.682171111) + 0.012522878);
}

vec4 GammaToLinearSpace (vec4 sRGB)
{
    return vec4(GammaToLinearSpace(sRGB.rgb), sRGB.a);
}


vec3 LinearToGammaSpace (vec3 linRGB)
{
    linRGB = max(linRGB, vec3(0.0, 0.0, 0.0));
    // An almost-perfect approximation from http://chilliant.blogspot.com.au/2012/08/srgb-approximations-for-hlsl.html?m=1
    return max(1.055 * pow(linRGB, vec3(0.416666667)) - 0.055, 0.0);
}

vec3 decodeHDR(vec4 rgbm)
{
    return 5.0 * rgbm.rgb * rgbm.a;
}

vec3 Environment(samplerCube env, vec3 reflUVW)
{
    //to match DX coordinate handedness, which inherit from DDS texture
    return decodeHDR(GammaToLinearSpace(textureCube(env, reflUVW.zyx)));
}

float decode_float_in_RGBA(vec4 rgba)
{
    return dot(rgba, vec4(1.0, 1.0 / 255.0, 1.0 / 65025.0, 1.0 / 16581375.0));
}

float linearize_depth(float depth)
{
    float z = depth * 2.0 - 1.0; // Back to NDC 
    return (2.0 * near_plane * far_plane) / (far_plane + near_plane - z * (far_plane - near_plane));
}

float calculate_shadow(vec4 frag_pos_light_space)
{
    vec3 projCoords = frag_pos_light_space.xyz / frag_pos_light_space.w;
    projCoords = projCoords * 0.5 + 0.5;
    float currentDepth = projCoords.z;
    float bias = 0.003;
    float shadow = 0.0;
    
    vec2 depth_map_size = 1.0 / vec2(1024, 1024);
    for (int x = -1; x <= 1; ++x)
    {
        for (int y = -1; y <= 1; ++y)
        {
            float pcf_depth = decode_float_in_RGBA(texture2D(depth_map, projCoords.xy + vec2(x, y) * depth_map_size));
            shadow += currentDepth - bias > pcf_depth ? 1.0 : 0.0;
        }
    }
    shadow /= 9.0;

    if (projCoords.z > 1.0)
    {
        shadow = 0.0;
    }
    
    return shadow;
}

vec4 Albedo(vec2 uv)
{
    return texture2D(diffuse_map, uv, -1.0);
}

vec3 calculate_directional_light(vec3 dir, vec3 normal, vec3 view_dir, vec3 frag_color)
{
    vec3 light_dir = normalize(-dir);
    // diffuse shading
    float diff = max(dot(normal, light_dir), 0.0);
    // specular shading
    vec3 halfway_dir = normalize(light_dir + view_dir);
    float spec = pow(max(dot(normal, halfway_dir), 0.0), 32.0);
    // combine results
    vec3 ambient = vec3(directional_light_ambient) * frag_color;
    vec3 diffuse = vec3(directional_light_diffuse) * diff * frag_color * var_spherical_harmonic_lighting;
    vec3 specular = vec3(directional_light_specular) * spec * frag_color;
    
    float shadow = calculate_shadow(var_frag_pos_light_space);
    return ambient + (1.0 - shadow) * (diffuse + specular);
}

vec3 calculate_point_light(vec3 light_pos, vec3 normal, vec3 frag_pos, vec3 view_dir, vec3 frag_color)
{
    vec3 light_dir = normalize(light_pos - frag_pos);
    // diffuse shading
    float diff = max(dot(normal, light_dir), 0.0);
    // specular shading
    vec3 halfway_dir = normalize(light_dir + view_dir);
    float spec = pow(max(dot(normal, halfway_dir), 0.0), 32.0);
    // attenuation
    float distance = length(light_pos - frag_pos);
    float constant = light_attenuation_params.x;
    float linear = light_attenuation_params.y;
    float quadratic = light_attenuation_params.z;
    float attenuation = 1.0 / (constant + linear * distance + quadratic * (distance * distance));
    // combine results
    vec3 ambient = vec3(point_light_ambient) * frag_color;
    vec3 diffuse = vec3(point_light_diffuse) * diff * frag_color * var_spherical_harmonic_lighting;
    vec3 specular = vec3(point_light_specular) * spec * frag_color;
    ambient = ambient * attenuation;
    diffuse = diffuse * attenuation;
    specular = specular * attenuation;
    return (ambient + diffuse + specular);
}

vec3 calculate_spot_light(vec3 light_pos, vec3 normal, vec3 frag_pos, vec3 view_dir ,vec3 frag_color)
{
    vec3 light_dir = normalize(light_pos - frag_pos);
    // diffuse shading
    float diff = max(dot(normal, light_dir), 0.0);
    // specular shading
    vec3 halfway_dir = normalize(light_dir + view_dir);
    float spec = pow(max(dot(normal, halfway_dir), 0.0), 32.0);
    // attenuation
    float distance = length(light_pos - frag_pos);
    float constant = light_attenuation_params.x;
    float linear = light_attenuation_params.y;
    float quadratic = light_attenuation_params.z;
    float attenuation = 1.0 / (constant + linear * distance + quadratic * (distance * distance));
    // spotlight intensitys
    float theta = dot(light_dir, normalize(vec3(light_pos.x, light_pos.y, light_pos.z)));
    float cut_off = spot_light_inout_cut_off.x;
    float outter_cut_off = spot_light_inout_cut_off.y;
    float epsilon = cut_off - outter_cut_off;
    float intensity = clamp((theta - outter_cut_off) / epsilon, 0.0, 1.0);
    // combine results
    vec3 ambient = vec3(spot_light_ambient) * frag_color;
    vec3 diffuse = vec3(spot_light_diffuse) * diff * frag_color * var_spherical_harmonic_lighting;
    vec3 specular = vec3(spot_light_specular) * spec * frag_color;
    //var_spherical_harmonic_lighting
    ambient = ambient * attenuation * intensity;
    diffuse = diffuse * attenuation * intensity;
    specular = specular * attenuation * intensity;

    return ambient + diffuse + specular;
}

vec3 calculate_env(vec3 view_dir, vec3 normal)
{
    vec3 R = reflect(view_dir, normal);
    return decodeHDR(GammaToLinearSpace(textureCube(env_map, R.zyx)));
}

void main()
{
    vec3 frag_color = vec3(Albedo(var_texcoord0.xy));

    vec3 normal = normalize(var_normal);
    vec3 view_dir = normalize(vec3(view_pos - var_frag_pos));

    vec3 color = calculate_directional_light(vec3(directional_light_dir), normal, view_dir, frag_color);

    color += calculate_point_light(vec3(point_light_pos), normal, vec3(var_frag_pos), view_dir, frag_color);

    color += calculate_spot_light(vec3(spot_light_pos), normal, vec3(var_frag_pos), view_dir, frag_color);

    color += LinearToGammaSpace(calculate_env(-view_dir, normal));

    gl_FragColor = vec4(color, 1.0) * tint;
}